﻿CompilerIf Defined(INCLUDE_STACKCLASS, #PB_Constant)=0
#INCLUDE_STACKCLASS=1
;/////////////////////////////////////////////////////////////////////////////////
;An OOP class for implementing either a simple stack or a structured stack.

;nxSoftware.
;January 2010.

;Developed with Purebasic 4.4.
;Fully cross-platform.
;/////////////////////////////////////////////////////////////////////////////////
;/////////////////////////////////////////////////////////////////////////////////
;To create an instance of a new stack class, use something like :
;
;   MyStackObject.StackObject
;   MyStackObject = NewStack(numberOfElements)
;
;and check the return value is non zero.
;
;For a structured stack use :
;
;   MyStackObject.StructuredStackObject
;   MyStackObject = NewStructuredStack(numElements, sizeOfEachElement, numberOfStringFields)
;
;and check the return value is non zero.
;
;NOTE: for a structured stack, ALL string fields must be placed before all other fields.
;/////////////////////////////////////////////////////////////////////////////////
;/////////////////////////////////////////////////////////////////////////////////
;Methods exposed by both of the stack classes :
;
;      Destroy()
;      NumberOfElementsPushed.i()
;      Peek.i(index)
;      Pop.i()
;      Push.i(value)
;      ResetStack()
;      ResizeStack.i(NumElements)
;      SizeOfStack.i()
;/////////////////////////////////////////////////////////////////////////////////


XIncludeFile "stackClass_Residents.pbi"

;The following structure contains the class template and private properties,
  Structure _membersStackClass
    *vTable
    NumElements.i
    StackBase.i
    StackPtr.i
    ;The following field(s) are used by structured stack objects.
      SizeOfEachElement.i
      NumberOfStringFields.i  ;All strng fields must reside at the beginning of the
                              ;structure.
  EndStructure 


Declare StructuredStackClass_Pop(*this._membersStackClass, ptrStructure=0)


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of the stackClass.
;Returns zero if an error.
Procedure.i NewStack(NumElements)
  Protected *object._membersStackClass, StackBase
  If NumElements > 0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersStackClass))
      If *object
        ;Allocate memory for the stack.
          StackBase = AllocateMemory(NumElements*SizeOf(INTEGER))
          If stackBase
            ;Set the object properties.
              *object\vTable = ?VTable_StackClass
              *object\NumElements = numElements
              *object\StackBase= stackBase
              *object\StackPtr = 0
          Else
            FreeMemory(*object)
            *object = 0          
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of a structured stackClass.
;Returns zero if an error.
Procedure.i NewStructuredStack(numElements, sizeOfEachElement, numberOfStringFields)
  Protected *object._membersStackClass, stackBase
  If numElements And sizeOfEachElement And numberOfStringFields>=0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersStackClass))
      If *object
        ;Allocate memory for the stack.
          stackBase = AllocateMemory(numElements*SizeOf(INTEGER))
          If stackBase
            ;Set the object properties.
              *object\vTable = ?VTable_StructuredStackClass
              *object\NumElements = numElements
              *object\StackBase= stackBase
              *object\StackPtr = 0
              *object\SizeOfEachElement = sizeOfEachElement
              *object\NumberOfStringFields = numberOfStringFields
          Else
            FreeMemory(*object)
            *object = 0          
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-BASE CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following returns the number of elements pushed onto the stack.
Procedure.i stackBaseClass_NumberOfElementsPushed(*this._membersStackClass)
  ProcedureReturn *this\StackPtr
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following returns the size of the stack.
Procedure.i stackBaseClass_SizeOfStack(*this._membersStackClass)
  ProcedureReturn *this\NumElements
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-BASIC STACK CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with the underlying stack.
Procedure stackClass_Destroy(*this._membersStackClass)
  FreeMemory(*this\StackBase)
  FreeMemory(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following returns the specified element from the underlying stack without
;performing any kind of 'pop'.
;'Index' = 0 or 1 or 2 or ... or NumberOfElementsPushed()-1.
Procedure.i stackClass_Peek(*this._membersStackClass, index)
  Protected *ptr.INTEGER
  If index>=0 And index < *this\StackPtr
  ;Good to go.
    *ptr = *this\StackBase + index*SizeOf(INTEGER)
    ProcedureReturn *ptr\i
EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following pops an element from the underlying stack.
Procedure.i stackClass_Pop(*this._membersStackClass)
  Protected *ptr.INTEGER
  If *this\StackPtr
    *this\StackPtr - 1
    *ptr = *this\StackBase + *this\StackPtr*SizeOf(INTEGER)
    ProcedureReturn *ptr\i
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following pushes a new element to the underlying stack if there is room.
;Returns zero if an error, i.e. if the stack is full.
Procedure.i stackClass_Push(*this._membersStackClass, Value)
  Protected *ptr.INTEGER
  If *this\StackPtr < *this\NumElements
    *ptr = *this\StackBase + *this\StackPtr*SizeOf(INTEGER)
    *ptr\i = Value
    *this\StackPtr + 1
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following resets a stack and effectively clears it's contents.
Procedure stackClass_ResetStack(*this._membersStackClass)
  *this\StackPtr = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following attempts to dynamically resize the underlying stack.
;If the stack is reduced in size then the stack pointer may be lowered in order
;to point at the top of the stack etc.
;Returns zero if an error.
Procedure.i stackClass_ResizeStack(*this._membersStackClass, numElements)
  Protected mem
  If numElements And *this\NumElements <> numElements
    mem = ReAllocateMemory(*this\StackBase, numElements*SizeOf(INTEGER))
    If mem ;Reallocation was successful.
      *this\NumElements = numElements
      *this\StackBase = mem
      ;Reset the stack pointer if required.
        If *this\StackPtr > numElements
          *this\StackPtr = numElements
        EndIf
    EndIf
  EndIf
  ProcedureReturn mem
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-STRUCTURED STACK CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with a structured stack.
Procedure StructuredStackClass_Destroy(*this._membersStackClass)
  Protected i, count
  ;First free all elements placed on the stack because each has memory reserved.
    count = *this\StackPtr
    For i = 1 To count
      StructuredStackClass_Pop(*this, 0)
    Next
  FreeMemory(*this\StackBase)
  FreeMemory(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following returns the specified element from the underlying stack without
;performing any kind of 'pop'.
;'Index' = 0 or 1 or 2 or ... or NumberOfElementsPushed()-1.
;Returns zero if an error.
Procedure.i StructuredStackClass_Peek(*this._membersStackClass, index, ptrStructure)
  Protected text$, *ptrStr.String, i, *ptrINTEGER.INTEGER, memStruct
  If ptrStructure And index>=0 And index < *this\StackPtr
  ;Good to go.
    *ptrINTEGER = *this\StackBase + index*SizeOf(INTEGER)
    memStruct = *ptrINTEGER\i  ;The address of the structure which was pushed onto the stack.
    ;Sort out any string fields.
      *ptrStr = ptrStructure : *ptrINTEGER = memStruct
      For i = 1 To *this\NumberOfStringFields
        If *ptrINTEGER\i
          text$ = PeekS(*ptrINTEGER\i)
          ;We need to copy this string to the structure specified by ptrStructure.
          ;We get PB to do this in such a way that the heap memory is allocated by PB itself
          ;so that it knows to free it as appropriate and also frees any existing string in the
          ;structure being overwritten.
          ;This means that we can then free the string originally pushed on the stack.
            *ptrStr\s = text$
        Else
          If PeekI(*ptrStr)
            *ptrStr\s = ""  ;This means any existing string will be freed by PB.
          EndIf
        EndIf
        *ptrStr + SizeOf(INTEGER)
        *ptrINTEGER+SizeOf(INTEGER)
      Next
    ;Copy the remaining fields; those following the string fields.
      CopyMemory(memStruct+*this\NumberOfStringFields*SizeOf(INTEGER), ptrStructure+*this\NumberOfStringFields*SizeOf(INTEGER), *this\SizeOfEachElement-*this\NumberOfStringFields*SizeOf(INTEGER))
    ProcedureReturn #True
  EndIf
  ProcedureReturn #False
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following pops a structured element from the underlying stack.
;If ptrStructure is zero then an element is popped and discarded (all appropriate memory
;is freed).
Procedure StructuredStackClass_Pop(*this._membersStackClass, ptrStructure=0)
  Protected text$, *ptrStr.String, i, *ptrINTEGER.INTEGER, memStruct
  If *this\StackPtr
    *this\StackPtr - 1
    *ptrINTEGER = *this\StackBase + *this\StackPtr*SizeOf(INTEGER)
    memStruct = *ptrINTEGER\i  ;The address of the structure which was pushed onto the stack.
    ;Sort out any string fields.
      *ptrStr = ptrStructure : *ptrINTEGER = memStruct
      For i = 1 To *this\NumberOfStringFields
        If *ptrINTEGER\i
          If ptrStructure
            text$ = PeekS(*ptrINTEGER\i)
            ;We need to copy this string to the structure specified by ptrStructure.
            ;We get PB to do this in such a way that the heap memory is allocated by PB itself
            ;so that it knows to free it as appropriate and also frees any existing string in the
            ;structure being overwritten.
            ;This means that we can then free the string originally pushed on the stack.
              *ptrStr\s = text$
          EndIf
          FreeMemory(*ptrINTEGER\i)
        Else
          If ptrStructure And PeekI(*ptrStr)
            *ptrStr\s = ""  ;This means any existing string will be freed by PB.
          EndIf
        EndIf
        *ptrStr + SizeOf(INTEGER)
        *ptrINTEGER+SizeOf(INTEGER)
      Next
    ;Copy the remaining fields; those following the string fields.
      If ptrStructure
        CopyMemory(memStruct+*this\NumberOfStringFields*SizeOf(INTEGER), ptrStructure+*this\NumberOfStringFields*SizeOf(INTEGER), *this\SizeOfEachElement-*this\NumberOfStringFields*SizeOf(INTEGER))
      EndIf
    ;Free the memory used to hold the pushed element.
      FreeMemory(memStruct)
    ProcedureReturn #True
  EndIf
  ProcedureReturn #False
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following pushes a new element to the underlying stack if there is room.
;ptrStructure is to contain the address of the structure to push.
;Returns zero if an error, e.g. if the stack is full.
Procedure.i StructuredStackClass_Push(*this._membersStackClass, ptrStructure)
  Protected i, blnError, *ptrStr.String, *ptrINTEGER.INTEGER, memStruct, memStr, text$
  If *this\StackPtr < *this\NumElements
  ;Allocate memory for the structure.
    memStruct = AllocateMemory(*this\SizeOfEachElement)
    If memStruct
      CopyMemory(ptrStructure, memStruct, *this\SizeOfEachElement)
      ;Sort out any string fields.
        *ptrStr = ptrStructure : *ptrINTEGER = memStruct
        For i = 1 To *this\NumberOfStringFields
          If PeekI(*ptrStr)
            ;Copy the string.
              text$ = *ptrStr\s
              *ptrINTEGER\i = AllocateMemory((Len(text$)+1)*SizeOf(CHARACTER))
              If *ptrINTEGER\i
                PokeS(*ptrINTEGER\i, text$)
              Else
                blnError = #True
                Break
              EndIf
          EndIf
          *ptrStr + SizeOf(INTEGER)
          *ptrINTEGER + SizeOf(INTEGER)
        Next
        If blnError
          ;We need to release all memory allocated for this element and return an error.
            While i>1
              i-1
              *ptrINTEGER - SizeOf(INTEGER)
              FreeMemory(*ptrINTEGER\i)
            Wend
            FreeMemory(memStruct)
            ProcedureReturn #False
        EndIf
        ;Now push the address of this memory block onto the stack.
          stackClass_Push(*this, memStruct)
        ProcedureReturn #True
    Else
      ProcedureReturn #False
    EndIf
  Else
    ProcedureReturn #False
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following resets a structured stack and effectively clears it's contents.
Procedure StructuredStackClass_ResetStack(*this._membersStackClass)
  Protected i, count
  ;First free all elements placed on the stack because each has memory reserved.
    count = *this\StackPtr
    For i = 1 To count
      StructuredStackClass_Pop(*this, 0)
    Next
  *this\StackPtr = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following attempts to dynamically resize the underlying structured stack.
;If the stack is reduced in size then the stack pointer may be lowered in order
;to point at the top of the stack etc.
;Returns zero if an error.
Procedure.i StructuredStackClass_ResizeStack(*this._membersStackClass, NumElements)
  Protected i, count, mem
  If NumElements And *this\NumElements <> NumElements
    ;First, in the case that the stack is being reduced in size we must free all memory
    ;associated with the elements being discarded,
      If NumElements < *this\NumElements
        count = *this\NumElements - NumElements
        For i = 1 To count
          StructuredStackClass_Pop(*this, 0)
        Next
      EndIf
    ;Now reduce the stack.
      ProcedureReturn stackClass_ResizeStack(*this, NumElements)
  EndIf
  ProcedureReturn #False
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VIRTUAL TABLES.

DataSection 
  VTable_StackClass: 
    Data.i @stackBaseClass_NumberOfElementsPushed() 
    Data.i @stackBaseClass_SizeOfStack() 
    Data.i @stackClass_Destroy() 
    Data.i @stackClass_Peek() 
    Data.i @stackClass_Pop() 
    Data.i @stackClass_Push() 
    Data.i @stackClass_ResetStack() 
    Data.i @stackClass_ResizeStack() 
  VTable_StructuredStackClass: 
    Data.i @stackBaseClass_NumberOfElementsPushed() 
    Data.i @stackBaseClass_SizeOfStack() 
    Data.i @StructuredStackClass_Destroy() 
    Data.i @StructuredStackClass_Peek()
    Data.i @StructuredStackClass_Pop() 
    Data.i @StructuredStackClass_Push() 
    Data.i @StructuredStackClass_ResetStack() 
    Data.i @StructuredStackClass_ResizeStack() 
EndDataSection 
CompilerEndIf

; IDE Options = PureBasic 5.21 LTS (Windows - x86)
; CursorPosition = 46
; FirstLine = 44
; Folding = ---